@@ -1,7 +1,10 @@
-from Acquisition import IAcquirer, aq_base, ImplicitAcquisitionWrapper
+from Acquisition import IAcquirer, aq_base, aq_inner, \
+                        ImplicitAcquisitionWrapper, aq_chain
 from ZODB.interfaces import IConnection
+from ZPublisher.BaseRequest import RequestContainer
 from persistent import IPersistent
 from zope.component import adapter, adapts
+from zope.app.component.hooks import getSite
 from zope.interface import implements, implementer
 from zope.app.keyreference.interfaces import IKeyReference, NotYet
 from zope.app.keyreference.persistent import KeyReferenceToPersistent
@@ -15,9 +18,9 @@ from zope.app.container.interfaces import IObjectAddedEvent
 def connectionOfPersistent(obj):
     """ zope2 cxn fetcher for wrapped items """
     cur = obj
-    if IAcquirer.providedBy(obj):
+    if IAcquirer.providedBy(obj) or hasattr(obj, '__parent__'):
         while not getattr(cur, '_p_jar', None):
-            cur = getattr(cur, 'aq_parent', None)
+            cur = getattr(cur, 'aq_parent', getattr(cur, '__parent__', None))
             if cur is None:
                 return None
         return cur._p_jar
@@ -46,7 +49,12 @@ class KeyReferenceToPersistent(KeyReferenceToPersistent):
     key_type_id = 'five.intid.keyreference'
 
     def __init__(self, wrapped_obj):
-        self.path = '/'.join(wrapped_obj.getPhysicalPath())
+        # make sure our object is wrapped by containment only
+        try:
+            wrapped_obj = aq_inner(wrapped_obj)
+            self.path = '/'.join(wrapped_obj.getPhysicalPath())
+        except AttributeError:
+            self.path = None
         self.object = aq_base(wrapped_obj)
         connection = IConnection(wrapped_obj, None)
 
@@ -60,7 +68,13 @@ class KeyReferenceToPersistent(KeyReferenceToPersistent):
                 raise UnsettableAttributeError(wrapped_obj)
             connection.add(self.object)
 
-        self.root_oid = get_root(wrapped_obj)._p_oid            
+        try:
+            self.root_oid = get_root(wrapped_obj)._p_oid
+        except AttributeError:
+            # If the object is unwrapped we can try to use the Site from the
+            # threadlocal as our acquisition context, hopefully it's not
+            # something odd.
+            self.root_oid = get_root(getSite())._p_oid
         self.oid = self.object._p_oid
         self.dbname = connection.db().database_name
 
@@ -70,11 +84,27 @@ class KeyReferenceToPersistent(KeyReferenceToPersistent):
 
     @property
     def wrapped_object(self):
-        return self.root.restrictedTraverse(self.path)
-    
+        if self.path is None:
+            return self.object
+        obj = self.root.restrictedTraverse(self.path)
+        chain = aq_chain(obj)
+        # Try to ensure we have a request at the acquisition root
+        # by using the one from getSite
+        if not len(chain) or not isinstance(chain[-1], RequestContainer):
+            site = getSite()
+            site_chain = aq_chain(site)
+            if len(site_chain) and isinstance(site_chain[-1], RequestContainer):
+                req = site_chain[-1]
+                new_obj = req
+                # rebuld the chain with the request at the bottom
+                for item in reversed(chain):
+                    new_obj = aq_base(item).__of__(new_obj)
+                obj = new_obj
+        return obj
+
     def __call__(self):
         return self.wrapped_object
-        
+
     def __hash__(self):
         return hash((self.dbname,
                      self.object._p_oid,
